@@ -54,10 +54,10 @@ const READ_ONLY_ANNO = "placesInternal/READ_ONLY";
 const LMANNO_FEEDURI = "livemark/feedURI";
 const LMANNO_SITEURI = "livemark/siteURI";
 
-//@line 62 "/export/home/mozilla/uild/firefox-3.0.4/src/mozilla/toolkit/components/places/src/utils.js"
+//@line 62 "/export/home/mozilla/packages/BUILD/SFWfirefox/mozilla/toolkit/components/places/src/utils.js"
 // On other platforms, the transferable system converts "\r\n" to "\n".
 const NEWLINE = "\r\n";
-//@line 65 "/export/home/mozilla/uild/firefox-3.0.4/src/mozilla/toolkit/components/places/src/utils.js"
+//@line 65 "/export/home/mozilla/packages/BUILD/SFWfirefox/mozilla/toolkit/components/places/src/utils.js"
 
 function QI_node(aNode, aIID) {
   var result = null;
@@ -1080,11 +1080,26 @@ var PlacesUtils = {
               // remove tags via the tagging service
               var tags = this._utils.tagging.allTags;
               var uris = [];
+              var bogusTagContainer = false;
               for (let i in tags) {
-                var tagURIs = this._utils.tagging.getURIsForTag(tags[i]);
+                var tagURIs = [];
+                // skip empty tags since getURIsForTag would throw
+                if (tags[i])
+                  tagURIs = this._utils.tagging.getURIsForTag(tags[i]);
+
+                if (!tagURIs.length) {
+                  // This is a bogus tag container, empty tags should be removed
+                  // automatically, but this does not work if they contain some
+                  // not-uri node, so we remove them manually.
+                  // XXX this is a temporary workaround until we implement
+                  // preventive database maintainance in bug 431558.
+                  bogusTagContainer = true;
+                }
                 for (let j in tagURIs)
                   this._utils.tagging.untagURI(tagURIs[j], [tags[i]]);
               }
+              if (bogusTagContainer)
+                this._utils.bookmarks.removeFolderChildren(rootItemId);
             }
             else if ([this._utils.toolbarFolderId,
                       this._utils.unfiledBookmarksFolderId,
@@ -1168,7 +1183,11 @@ var PlacesUtils = {
         if (aContainer == PlacesUtils.bookmarks.tagsFolder) {
           if (aData.children) {
             aData.children.forEach(function(aChild) {
-              this.tagging.tagURI(this._uri(aChild.uri), [aData.title]);
+              try {
+                this.tagging.tagURI(this._uri(aChild.uri), [aData.title]);
+              } catch (ex) {
+                // invalid tag child, skip it
+              }
             }, this);
             return [folderIdMap, searchIds];
           }
@@ -1426,9 +1445,9 @@ var PlacesUtils = {
           var childNode = aSourceNode.getChild(i);
           if (aExcludeItems && aExcludeItems.indexOf(childNode.itemId) != -1)
             continue;
-          if (i != 0)
+          var written = serializeNodeToJSONStream(aSourceNode.getChild(i), i);
+          if (written && i < cc - 1)
             aStream.write(",", 1);
-          serializeNodeToJSONStream(aSourceNode.getChild(i), i);
         }
         if (!wasOpen)
           aSourceNode.containerOpen = false;
@@ -1449,24 +1468,53 @@ var PlacesUtils = {
 
       addGenericProperties(bNode, node);
 
-      if (self.nodeIsURI(bNode))
+      var parent = bNode.parent;
+      var grandParent = parent ? parent.parent : null;
+
+      if (self.nodeIsURI(bNode)) {
+        // Tag root accept only folder nodes
+        if (parent && parent.itemId == self.tagsFolderId)
+          return false;
+        // Check for url validity, since we can't halt while writing a backup.
+        // This will throw if we try to serialize an invalid url and it does
+        // not make sense saving a wrong or corrupt uri node.
+        try {
+          self._uri(bNode.uri);
+        } catch (ex) {
+          return false;
+        }
         addURIProperties(bNode, node);
-      else if (self.nodeIsContainer(bNode))
+      }
+      else if (self.nodeIsContainer(bNode)) {
+        // Tag containers accept only uri nodes
+        if (grandParent && grandParent.itemId == self.tagsFolderId)
+          return false;
         addContainerProperties(bNode, node);
-      else if (self.nodeIsSeparator(bNode))
+      }
+      else if (self.nodeIsSeparator(bNode)) {
+        // Tag root accept only folder nodes
+        // Tag containers accept only uri nodes
+        if ((parent && parent.itemId == self.tagsFolderId) ||
+            (grandParent && grandParent.itemId == self.tagsFolderId))
+          return false;
+
         addSeparatorProperties(bNode, node);
+      }
 
       if (!node.feedURI && node.type == self.TYPE_X_MOZ_PLACE_CONTAINER)
         writeComplexNode(aStream, node, bNode);
       else
         writeScalarNode(aStream, node);
+      return true;
     }
 
     // serialize to stream
     serializeNodeToJSONStream(aNode, null);
   },
 
-  // XXX testing serializers
+  /**
+   * Serialize a JS object to JSON
+   */
   toJSONString: function PU_toJSONString(aObj) {
     var JSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON);
     return JSON.encode(aObj);
@@ -1514,6 +1562,23 @@ var PlacesUtils = {
     stream.close();
   },
 
+  /**
+   * Creates a filename for bookmarks backup files.
+   *
+   * @param [optional] aDateObj Date object used to build the filename.
+   *                            Will use current date if empty.
+   * @return A bookmarks backup filename.
+   */
+  getBackupFilename:
+  function PU_getBackupFilename(aDateObj) {
+    if (!aDateObj)
+      aDateObj = new Date();
+    // Use YYYY-MM-DD (ISO 8601) as it doesn't contain illegal characters
+    // and makes the alphabetical order of multiple backup files more useful.
+    var date = aDateObj.toLocaleFormat("%Y-%m-%d");
+    return "bookmarks-" + date + ".json";
+  },
+
   /**
    * ArchiveBookmarksFile()
    *
@@ -1538,24 +1603,31 @@ var PlacesUtils = {
         return; // unable to create directory!
     }
 
-    // construct the new leafname
-    // Use YYYY-MM-DD (ISO 8601) as it doesn't contain illegal characters
-    // and makes the alphabetical order of multiple backup files more useful.
-    var date = new Date().toLocaleFormat("%Y-%m-%d");
-    var backupFilename = this.getFormattedString("bookmarksArchiveFilename", [date]);
-
+    // Construct the new leafname.
+    var date = new Date();
+    var backupFilename = this.getBackupFilename(date);
     var backupFile = null;
     if (!aForceArchive) {
       var backupFileNames = [];
       var backupFilenamePrefix = backupFilename.substr(0, backupFilename.indexOf("-"));
+
+      // Get the localized backup filename, to clear out
+      // old backups with a localized name (bug 445704).
+      var localizedFilename = this.getFormattedString("bookmarksArchiveFilename", [date]);
+      var localizedFilenamePrefix = localizedFilename.substr(0, localizedFilename.indexOf("-"));
+      var rx = new RegExp("^(bookmarks|" + localizedFilenamePrefix + ")-([0-9-]+)\.(json|html)");
+
       var entries = bookmarksBackupDir.directoryEntries;
       while (entries.hasMoreElements()) {
         var entry = entries.getNext().QueryInterface(Ci.nsIFile);
         var backupName = entry.leafName;
-        if (backupName.substr(0, backupFilenamePrefix.length) == backupFilenamePrefix) {
+        // A valid backup is any file that matches either the localized or
+        // not-localized filename (bug 445704).
+        var matches = backupName.match(rx);
+        if (matches) {
           if (backupName == backupFilename)
             backupFile = entry;
-          backupFileNames.push(backupName);
+          backupFileNames.push({ filename: backupName, date: matches[2] });
         }
       }
 
@@ -1569,11 +1641,12 @@ var PlacesUtils = {
         // number specified in the pref.
         if (!backupFile)
           numberOfBackupsToDelete++;
-
-        backupFileNames.sort();
+        backupFileNames.sort(function compare(a, b) {
+          return a.date < b.date ? -1 : a.date > b.date ? 1 : 0;
+        });
         while (numberOfBackupsToDelete--) {
           let backupFile = bookmarksBackupDir.clone();
-          backupFile.append(backupFileNames[0]);
+          backupFile.append(backupFileNames[0].filename);
           backupFile.remove(false);
           backupFileNames.shift();
         }
